@@ -5,14 +5,13 @@
 use pyo3::prelude::*;
 
 #[macro_use]
-#[path = "../src/tests/common.rs"]
-mod common;
+mod test_utils;
 
 macro_rules! make_struct_using_macro {
     // Ensure that one doesn't need to fall back on the escape type: tt
     // in order to macro create pyclass.
     ($class_name:ident, $py_name:literal) => {
-        #[pyclass(name=$py_name)]
+        #[pyclass(name=$py_name, subclass)]
         struct $class_name {}
     };
 }
@@ -31,7 +30,7 @@ macro_rules! set_extends_via_macro {
 set_extends_via_macro!(MyClass2, MyBaseClass);
 
 //
-// Check that pyfunctiona nd text_signature can be called with macro arguments.
+// Check that pyfunctions and text_signature can be called with macro arguments.
 //
 
 macro_rules! fn_macro {
@@ -73,18 +72,18 @@ property_rename_via_macro!(my_new_property_name);
 
 #[test]
 fn test_macro_rules_interactions() {
-    Python::with_gil(|py| {
-        let my_base = py.get_type_bound::<MyBaseClass>();
+    Python::attach(|py| {
+        let my_base = py.get_type::<MyBaseClass>();
         py_assert!(py, my_base, "my_base.__name__ == 'MyClass'");
 
-        let my_func = wrap_pyfunction_bound!(my_function_in_macro, py).unwrap();
+        let my_func = wrap_pyfunction!(my_function_in_macro, py).unwrap();
         py_assert!(
             py,
             my_func,
             "my_func.__text_signature__ == '(a, b=None, *, c=42)'"
         );
 
-        let renamed_prop = py.get_type_bound::<ClassWithProperty>();
+        let renamed_prop = py.get_type::<ClassWithProperty>();
         py_assert!(
             py,
             renamed_prop,
