@@ -1,4 +1,4 @@
-#![cfg(all(feature = "macros", not(PyPy)))]
+#![cfg(all(feature = "macros", not(any(PyPy, GraalPy))))]
 
 use pyo3::{prelude::*, types::PySuper};
 
@@ -35,16 +35,15 @@ impl SubClass {
     }
 
     fn method_super_new<'py>(self_: &Bound<'py, Self>) -> PyResult<Bound<'py, PyAny>> {
-        #[cfg_attr(not(feature = "gil-refs"), allow(deprecated))]
-        let super_ = PySuper::new_bound(&self_.get_type(), self_)?;
+        let super_ = PySuper::new(&self_.get_type(), self_)?;
         super_.call_method("method", (), None)
     }
 }
 
 #[test]
 fn test_call_super_method() {
-    Python::with_gil(|py| {
-        let cls = py.get_type_bound::<SubClass>();
+    Python::attach(|py| {
+        let cls = py.get_type::<SubClass>();
         pyo3::py_run!(
             py,
             cls,
